@@ -50,6 +50,7 @@
 #'   be set in the global option \code{tinytex.engine_args}, e.g.,
 #'   \code{options(tinytex.engine_args = '-shell-escape'}).
 #' @param emulation Whether to emulate the executable \command{latexmk} using R.
+#'   Note that this is unused when \code{engine == 'tectonic'}.
 #' @param min_times,max_times The minimum and maximum number of times to rerun
 #'   the LaTeX engine when using emulation. You can set the global options
 #'   \code{tinytex.compile.min_times} or \code{tinytex.compile.max_times}, e.g.,
@@ -69,7 +70,7 @@
 #' @return A character string of the path of the output file (i.e., the value of
 #'   the \code{pdf_file} argument).
 latexmk = function(
-  file, engine = c('pdflatex', 'xelatex', 'lualatex', 'latex'),
+  file, engine = c('pdflatex', 'xelatex', 'lualatex', 'latex', 'tectonic'),
   bib_engine = c('bibtex', 'biber'), engine_args = NULL, emulation = TRUE,
   min_times = 1, max_times = 10, install_packages = emulation && tlmgr_writable(),
   pdf_file = gsub('tex$', 'pdf', file), clean = TRUE
@@ -115,6 +116,10 @@ latexmk = function(
     file_rename(pdf, pdf_file)
     pdf_file
   }
+  if (engine == 'tectonic') {
+    system2_quiet('tectonic', c(engine_args, shQuote(file)))
+    return(check_pdf())
+  }
   if (emulation) {
     latexmk_emu(
       file, engine, bib_engine, engine_args, min_times, max_times,
@@ -161,10 +166,11 @@ latexmk_emu = function(
     'nav', 'snm', 'vrb', 'ilg', 'ind', 'xwm', 'brf', 'run.xml'
   )
   base = gsub('[.]tex$', '', basename(file))
-  aux_files = setNames(paste(base, aux, sep = '.'), aux)
+  aux_files = paste(base, aux, sep = '.')
   aux_files = c(aux_files, 'preview.aux')  # generated by the preview package
   if (!is.null(output_dir <- getOption('tinytex.output_dir')))
     aux_files = file.path(output_dir, aux_files)
+  names(aux_files)[seq_along(aux)] = aux
   logfile = aux_files['log']; unlink(logfile)  # clean up the log before compilation
 
   # clean up aux files from LaTeX compilation
@@ -213,17 +219,20 @@ latexmk_emu = function(
         on_error()
       }, logfile = logfile, fail_rerun = verbose
     )
-    # PNAS you are the worst! Why don't you singal an error in case of missing packages?
+    # PNAS you are the worst! Why don't you signal an error in case of missing packages?
     if (res == 0 && !file.exists(filep)) on_error()
     invisible(res)
   }
   run_engine()
-  if (install_packages) check_babel(logfile)
+  # some problems only trigger warnings but not errors, e.g.,
+  # https://github.com/yihui/tinytex/issues/311 fix them and re-run engine
+  if (install_packages && check_extra(logfile)) run_engine()
 
   # generate index
   if (file.exists(idx <- aux_files['idx'])) {
     idx_engine = getOption('tinytex.makeindex', 'makeindex')
     install_cmd(idx_engine)
+    run_engine()  # run the engine one more time (cf rstudio/bookdown#1274)
     system2_quiet(idx_engine, c(getOption('tinytex.makeindex.args'), shQuote(idx)), error = {
       stop("Failed to build the index via ", idx_engine, call. = FALSE)
     })
@@ -250,8 +259,7 @@ latexmk_emu = function(
           }
           if (!tlmgr_available() || !install_packages) return(warn())
           # install the possibly missing .bst package and rebuild bib
-          r = '.* open style file ([^ ]+).*'
-          pkgs = parse_packages(files = grep_sub(r, '\\1', x), quiet = !verbose)
+          pkgs = parse_packages(text = x, quiet = !verbose)
           if (length(pkgs) == 0 || identical(pkgs, pkgs_last)) return(warn())
           pkgs_last <<- pkgs
           tlmgr_install(pkgs); build_bib()
@@ -290,9 +298,9 @@ tweak_aux = function(aux, x = readLines(aux)) {
   writeLines(x, aux)
 }
 
-needs_rerun = function(log) {
+needs_rerun = function(log, text = xfun::read_utf8(log)) {
   any(grepl(
-    '(Rerun to get |Please \\(re\\)run | Rerun LaTeX\\.)', readLines(log),
+    '(Rerun to get |Please \\(?re\\)?run | Rerun LaTeX\\.)', text,
     useBytes = TRUE
   ))
 }
@@ -378,6 +386,8 @@ check_unicode = function(x) {
 # whether a LaTeX log file contains LaTeX or package (e.g. babel) warnings
 latex_warning = function(file, show = getOption('tinytex.latexmk.warning', TRUE)) {
   if (!file.exists(file)) return()
+  # if the option tinytex.latexmk.warning = FALSE, delete the log in latexmk_emu()
+  if (!show && missing(show)) return()
   x = readLines(file, warn = FALSE)
   if (length(i <- grep('^(LaTeX|Package [[:alnum:]]+) Warning:', x)) == 0) return()
   # these warnings may be okay (our Pandoc LaTeX template in rmarkdown may need an update)
@@ -390,17 +400,56 @@ latex_warning = function(file, show = getOption('tinytex.latexmk.warning', TRUE)
     j:n
   }))
   i = sort(unique(i))
-  if (show) warning(paste(x[i], collapse = '\n'), call. = FALSE)
+  if (show) for(msg in x[i]) warning(msg, call. = FALSE, immediate. = TRUE)
   x[i]
 }
 
-# check if any babel packages are missing
-check_babel = function(file) {
-  if (length(m <- latex_warning(file, FALSE)) == 0 || length(grep('^Package babel Warning:', m)) == 0)
-    return()
-  r = "^\\(babel).* language `([[:alpha:]]+)'.*$"
-  if (length(m <- grep_sub(r, '\\1', m)) == 0) return()
-  tlmgr_install(paste0('hyphen-', tolower(m)))
+# check if any babel/glossaries/... packages are missing
+check_extra = function(file) {
+  length(m <- latex_warning(file, FALSE)) > 0 &&
+    length(grep('^Package ([^ ]+) Warning:', m)) > 0 &&
+    any(
+      check_babel(m),
+      check_glossaries(m),
+      check_datetime2(m),
+      check_polyglossia(m)
+    )
+}
+
+check_babel = function(text) {
+  r = "^(\\(babel\\).* |Package babel Warning: No hyphenation patterns were preloaded for the )language [`']([^']+)'.*$"
+  if (length(m <- grep_sub(r, 'hyphen-\\2', text)) == 0) return(FALSE)
+  # (babel) the language `German (new orthography)' into the format
+  m = gsub('\\s.*', '', m)
+  tlmgr_install(tolower(m)) == 0
+}
+
+# Package glossaries Warning: No language module detected for `english'.
+# (glossaries)                Language modules need to be installed separately.
+# (glossaries)                Please check on CTAN for a bundle called
+# (glossaries)                `glossaries-english' or similar.
+check_glossaries = function(text) {
+  r = "^\\(glossaries).* [`']([^']+)'.*$"
+  if (length(m <- grep_sub(r, '\\1', text)) == 0) return(FALSE)
+  tlmgr_install(m) == 0
+}
+
+# Package polyglossia Warning: No hyphenation patterns were loaded for `hungarian'
+# Package polyglossia Warning: No hyphenation patterns were loaded for British English
+check_polyglossia = function(text) {
+  r = "^Package polyglossia Warning: No hyphenation patterns were loaded for ([`'][^']+'|British English).*"
+  if (length(m <- grep_sub(r, '\\1', text)) == 0) return(FALSE)
+  m[m == 'British English'] = 'english'
+  m = gsub("[`']", '', m)
+  tlmgr_install(paste0('hyphen-', m)) == 0
+}
+
+# Package datetime2 Warning: Date-Time Language Module `english' not installed on
+# input line xxx.
+check_datetime2 = function(text) {
+  r = "^Package datetime2 Warning: Date-Time Language Module [`']([^']+)' not installed.*$"
+  if (length(m <- grep_sub(r, '\\1', text)) == 0) return(FALSE)
+  tlmgr_install(paste0('datetime2-', m)) == 0
 }
 
 # check the version of latexmk
@@ -490,8 +539,7 @@ parse_packages = function(
   unique(pkgs)
 }
 
-# find filenames (could also be font names) from LaTeX error logs
-detect_files = function(text) {
+regex_errors = function() {
   # possible errors are like:
   # ! LaTeX Error: File `framed.sty' not found.
   # /usr/local/bin/mktexpk: line 123: mf: command not found
@@ -510,35 +558,64 @@ detect_files = function(text) {
   # ! Package tikz Error: I did not find the tikz library 'hobby'... named tikzlibraryhobby.code.tex
   # support file `supp-pdf.mkii' (supp-pdf.tex) is missing
   # ! I can't find file `hyph-de-1901.ec.tex'.
-  r = c(
-    ".*! Font [^=]+=([^ ]+).+ not loadable.*",
-    '.*! .*The font "([^"]+)" cannot be found.*',
-    '.*!.+ error:.+\\(file ([^)]+)\\): .*',
-    '.*Package widetext error: Install the ([^ ]+) package.*',
-    '.*Unable to find TFM file "([^"]+)".*',
-    # the above are messages about missing fonts; below are typically missing .sty or commands
-
-    ".* File `(.+eps-converted-to.pdf)'.*",
-    ".*xdvipdfmx:fatal: pdf_ref_obj.*",
-
-    '.* (tikzlibrary[^ ]+?[.]code[.]tex).*',
-
-    ".*! LaTeX Error: File `([^']+)' not found.*",
-    ".* file ['`]?([^' ]+)'? not found.*",
-    '.*the language definition file ([^ ]+) .*',
-    '.* \\(file ([^)]+)\\): cannot open .*',
-    ".*file `([^']+)' .*is missing.*",
-    ".*! CTeX fontset `([^']+)' is unavailable.*",
-    ".*: ([^:]+): command not found.*",
-    ".*! I can't find file `([^']+)'.*"
+  # ! Package pdfx Error: No color profile sRGB_IEC61966-2-1_black_scaled.icc found
+  # No file LGRcmr.fd. ! LaTeX Error: This NFSS system isn't set up properly.
+  list(
+    font = c(
+      # error messages about missing fonts (don't move the first item below, as
+      # it is special and emitted by widetext; the rest can be freely reordered)
+      '.*Package widetext error: Install the ([^ ]+) package.*',
+      ".*! Font [^=]+=([^ ]+).+ not loadable.*",
+      '.*! .*The font "([^"]+)" cannot be found.*',
+      '.*!.+ error:.+\\(file ([^)]+)\\): .*',
+      '.*Unable to find TFM file "([^"]+)".*'
+    ),
+    fd = c(
+      # font definition files
+      ".*No file ([^`'. ]+[.]fd)[.].*"
+    ),
+    epstopdf = c(
+      # possible errors when epstopdf is missing
+      ".* File [`'](.+eps-converted-to.pdf)'.*",
+      ".*xdvipdfmx:fatal: pdf_ref_obj.*"
+    ),
+    colorprofiles.sty = c(
+      '.* Package pdfx Error: No color profile ([^ ]+).*'
+    ),
+    tikz = c(
+      # when a required tikz library is missing
+      '.* (tikzlibrary[^ ]+?[.]code[.]tex).*'
+    ),
+    style = c(
+      # missing .sty or commands
+      ".* Loading '([^']+)' aborted!",
+      ".*! LaTeX Error: File [`']([^']+)' not found.*",
+      ".* file ['`]?([^' ]+)'? not found.*",
+      '.*the language definition file ([^ ]+) .*',
+      '.* \\(file ([^)]+)\\): cannot open .*',
+      '.* open style file ([^ ]+).*',
+      ".*file [`']([^']+)' .*is missing.*",
+      ".*! CTeX fontset [`']([^']+)' is unavailable.*",
+      ".*: ([^:]+): command not found.*",
+      ".*! I can't find file [`']([^']+)'.*"
+    )
   )
-  x = grep(paste(r, collapse = '|'), text, value = TRUE)
-  if (length(x) > 0) unique(unlist(lapply(r, function(p) {
+}
+
+# find filenames (could also be font names) from LaTeX error logs
+detect_files = function(text) {
+  r = regex_errors()
+  x = grep(paste(unlist(r), collapse = '|'), text, value = TRUE)
+  if (length(x) > 0) unique(unlist(lapply(unlist(r), function(p) {
     v = grep_sub(p, '\\1', x)
     if (length(v) == 0) return(v)
-    if (p == r[8] && length(grep('! Package tikz Error:', text)) == 0) return()
-    if (!(p %in% r[1:5])) return(if (p %in% r[6:7]) 'epstopdf' else v)
-    if (p == r[4]) paste0(v, '.sty') else font_ext(v)
+    if (p == r$tikz && length(grep('! Package tikz Error:', text)) == 0) return()
+    for (i in c('epstopdf', 'colorprofiles.sty')) {
+      if (p %in% r[[i]]) return(i)
+    }
+    if (p == r$fd) v = tolower(v)  # LGRcmr.fd -> lgrcmr.fd
+    if (!(p %in% r$font)) return(v)
+    if (p == r$font[1]) paste0(v, '.sty') else font_ext(v)
   })))
 }
 
